version_info = (6, 5, 4)
pre_info = ""
dev_info = ""


def create_valid_version(release_info, epoch=None, pre_input="", dev_input=""):
    """
    Creates a pep440 valid version of version number given a tuple integers
    and optional epoch, prerelease and developmental info.

    Parameters
    ----------
    release_info : Tuple(Int)
    epoch : Int, default None
    pre_input : Str, default ''
    dev_input : Str, default ''
    """

    pep440_err = "The version number is not a pep 440 compliant version number"

    if epoch is not None:
        epoch_seg = str(epoch) + "!"
    else:
        epoch_seg = ""

    release_seg = ".".join(map(str, release_info))

    _magic_pre = ["a", "b", "rc"]
    if pre_input != "" and not any([pre_input.startswith(prefix) for prefix in _magic_pre]):
        raise ValueError(pep440_err + "\n please fix your prerelease segment.")
    else:
        pre_seg = pre_input

    if dev_input == "":
        dev_seg = dev_input
    elif not dev_input.startswith(".") and dev_input.startswith("dev"):
        dev_seg = "".join([".", dev_input])
    elif dev_input.startswith(".dev"):
        dev_seg = dev_input
    elif dev_input != "":
        raise ValueError(pep440_err + "\n please fix your development segment.")

    if dev_input != "" and not any([dev_seg.endswith(str(n)) for n in range(10)]):
        dev_seg = "".join([dev_seg, "0"])

    out_version = "".join([epoch_seg, release_seg, pre_seg, dev_seg])

    import re

    def is_canonical(version):
        return (
            re.match(
                r"^([1-9]\d*!)?(0|[1-9]\d*)"
                r"(\.(0|[1-9]\d*))*((a|b|rc)(0|[1-9]\d*))?"
                r"(\.post(0|[1-9]\d*))?(\.dev(0|[1-9]\d*))?$",
                version,
            )
            is not None
        )

    if is_canonical(out_version):
        return out_version
    else:
        raise ValueError(pep440_err)


__version__ = create_valid_version(version_info, pre_input=pre_info, dev_input=dev_info)



"""
Module with tests for the RegexRemovePreprocessor.
"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import re

from nbformat import v4 as nbformat

from ..regexremove import RegexRemovePreprocessor
from .base import PreprocessorTestsBase


class TestRegexRemove(PreprocessorTestsBase):
    """Contains test functions for regexremove.py"""

    def build_notebook(self):
        notebook = super().build_notebook()
        # Add a few empty cells
        notebook.cells.extend(
            [
                nbformat.new_code_cell(""),
                nbformat.new_markdown_cell(" "),
                nbformat.new_raw_cell("\n"),
                nbformat.new_raw_cell("\t"),
            ]
        )

        return notebook

    def build_preprocessor(self):
        """Make an instance of a preprocessor"""
        preprocessor = RegexRemovePreprocessor()
        preprocessor.enabled = True
        return preprocessor

    def test_constructor(self):
        """Can a RegexRemovePreprocessor be constructed?"""
        self.build_preprocessor()

    def test_output(self):
        """Test the output of the RegexRemovePreprocessor"""
        pattern_lookup = {
            "disallow_whitespace": [r"\s*\Z"],
            "disallow_tab_newline": [r"\t\Z", r"\n\Z"],
        }
        expected_cell_count = {
            "default": 6,  # nothing is removed
            "disallow_whitespace": 2,  # all "empty" cells are removed
            "disallow_tab_newline": 4,  # cells with tab and newline are removed
            "none": 6,
        }
        for method in ["default", "disallow_whitespace", "disallow_tab_newline", "none"]:
            nb = self.build_notebook()
            res = self.build_resources()

            # Build the preprocessor and extend the list of patterns or use an empty list
            preprocessor = self.build_preprocessor()
            if method == "none":
                preprocessor.patterns = []
            else:
                preprocessor.patterns.extend(pattern_lookup.get(method, []))
            nb, res = preprocessor(nb, res)

            self.assertEqual(len(nb.cells), expected_cell_count[method])

            # Make sure none of the cells match the pattern
            patterns = list(map(re.compile, preprocessor.patterns))
            for cell in nb.cells:
                for pattern in patterns:
                    self.assertFalse(pattern.match(cell.source))



"""
Module with tests for files
"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import os

from ...tests.base import TestsBase
from ..files import FilesWriter


class Testfiles(TestsBase):
    """Contains test functions for files.py"""

    def test_basic_output(self):
        """Is FilesWriter basic output correct?"""

        # Work in a temporary directory.
        with self.create_temp_cwd():

            # Create the resoruces dictionary
            res = {}

            # Create files writer, test output
            writer = FilesWriter()
            writer.write("y", res, notebook_name="z")

            # Check the output of the file
            with open("z") as f:
                output = f.read()
                self.assertEqual(output, "y")

    def test_ext(self):
        """Does the FilesWriter add the correct extension to the output?"""

        # Work in a temporary directory.
        with self.create_temp_cwd():

            # Create the resoruces dictionary
            res = {"output_extension": ".txt"}

            # Create files writer, test output
            writer = FilesWriter()
            writer.write("y", res, notebook_name="z")

            # Check the output of the file
            assert os.path.isfile("z.txt")
            with open("z.txt") as f:
                output = f.read()
                self.assertEqual(output, "y")

    def test_extract(self):
        """Can FilesWriter write extracted figures correctly?"""

        # Work in a temporary directory.
        with self.create_temp_cwd():

            # Create the resoruces dictionary
            res = {"outputs": {os.path.join("z_files", "a"): b"b"}}

            # Create files writer, test output
            writer = FilesWriter()
            writer.write("y", res, notebook_name="z")

            # Check the output of the file
            with open("z") as f:
                output = f.read()
                self.assertEqual(output, "y")

            # Check the output of the extracted file
            extracted_file_dest = os.path.join("z_files", "a")
            assert os.path.isfile(extracted_file_dest)
            with open(extracted_file_dest) as f:
                output = f.read()
                self.assertEqual(output, "b")

    def test_build_dir(self):
        """Can FilesWriter write to a build dir correctly?"""

        # Work in a temporary directory.
        with self.create_temp_cwd():

            # Create the resoruces dictionary
            res = {"outputs": {os.path.join("z_files", "a"): b"b"}}

            # Create files writer, test output
            writer = FilesWriter()
            writer.build_directory = "build"
            writer.write("y", res, notebook_name="z")

            # Check the output of the file
            assert os.path.isdir(writer.build_directory)
            dest = os.path.join(writer.build_directory, "z")
            with open(dest) as f:
                output = f.read()
                self.assertEqual(output, "y")

            # Check the output of the extracted file
            extracted_file_dest = os.path.join(writer.build_directory, "z_files", "a")
            assert os.path.isfile(extracted_file_dest)
            with open(extracted_file_dest) as f:
                output = f.read()
                self.assertEqual(output, "b")

    def test_build_dir_default(self):
        """FilesWriter defaults to input path"""
        with self.create_temp_cwd():
            os.mkdir("sub")
            resources = {"metadata": {"path": "sub"}}
            writer = FilesWriter()
            writer.write("content", resources, notebook_name="out")
            dest = os.path.join("sub", "out")
            assert os.path.isfile(dest)
            with open(dest) as f:
                self.assertEqual(f.read().strip(), "content")

    def test_links(self):
        """Can the FilesWriter handle linked files correctly?"""

        # Work in a temporary directory.
        with self.create_temp_cwd():

            # Create test file
            os.mkdir("sub")
            with open(os.path.join("sub", "c"), "w") as f:
                f.write("d")

            # Create the resoruces dictionary
            res = {}

            # Create files writer, test output
            writer = FilesWriter()
            writer.files = [os.path.join("sub", "c")]
            writer.build_directory = "build"
            writer.write("y", res, notebook_name="z")

            # Check the output of the file
            assert os.path.isdir(writer.build_directory)
            dest = os.path.join(writer.build_directory, "z")
            with open(dest) as f:
                output = f.read()
                self.assertEqual(output, "y")

            # Check to make sure the linked file was copied
            path = os.path.join(writer.build_directory, "sub")
            assert os.path.isdir(path)
            dest = os.path.join(path, "c")
            assert os.path.isfile(dest)
            with open(dest) as f:
                output = f.read()
                self.assertEqual(output, "d")

    def test_glob(self):
        """Can the FilesWriter handle globbed files correctly?"""

        # Work in a temporary directory.
        with self.create_temp_cwd():

            # Create test files
            os.mkdir("sub")
            with open(os.path.join("sub", "c"), "w") as f:
                f.write("e")
            with open(os.path.join("sub", "d"), "w") as f:
                f.write("e")

            # Create the resoruces dictionary
            res = {}

            # Create files writer, test output
            writer = FilesWriter()
            writer.files = ["sub/*"]
            writer.build_directory = "build"
            writer.write("y", res, notebook_name="z")

            # Check the output of the file
            assert os.path.isdir(writer.build_directory)
            dest = os.path.join(writer.build_directory, "z")
            with open(dest) as f:
                output = f.read()
                self.assertEqual(output, "y")

            # Check to make sure the globbed files were copied
            path = os.path.join(writer.build_directory, "sub")
            assert os.path.isdir(path)
            for filename in ["c", "d"]:
                dest = os.path.join(path, filename)
                assert os.path.isfile(dest)
                with open(dest) as f:
                    output = f.read()
                    self.assertEqual(output, "e")

    def test_relpath(self):
        """Can the FilesWriter handle relative paths for linked files correctly?"""

        # Work in a temporary directory.
        with self.create_temp_cwd():

            # Create test file
            os.mkdir("sub")
            with open(os.path.join("sub", "c"), "w") as f:
                f.write("d")

            # Create the resoruces dictionary
            res = {}

            # Create files writer, test output
            writer = FilesWriter()
            writer.files = [os.path.join("sub", "c")]
            writer.build_directory = "build"
            writer.relpath = "sub"
            writer.write("y", res, notebook_name="z")

            # Check the output of the file
            assert os.path.isdir(writer.build_directory)
            dest = os.path.join(writer.build_directory, "z")
            with open(dest) as f:
                output = f.read()
                self.assertEqual(output, "y")

            # Check to make sure the linked file was copied
            dest = os.path.join(writer.build_directory, "c")
            assert os.path.isfile(dest)
            with open(dest) as f:
                output = f.read()
                self.assertEqual(output, "d")

    def test_relpath_default(self):
        """Is the FilesWriter default relative path correct?"""

        # Work in a temporary directory.
        with self.create_temp_cwd():

            # Create test file
            os.mkdir("sub")
            with open(os.path.join("sub", "c"), "w") as f:
                f.write("d")

            # Create the resoruces dictionary
            res = dict(metadata=dict(path="sub"))

            # Create files writer, test output
            writer = FilesWriter()
            writer.files = [os.path.join("sub", "c")]
            writer.build_directory = "build"
            writer.write("y", res, notebook_name="z")

            # Check the output of the file
            assert os.path.isdir(writer.build_directory)
            dest = os.path.join(writer.build_directory, "z")
            with open(dest) as f:
                output = f.read()
                self.assertEqual(output, "y")

            # Check to make sure the linked file was copied
            dest = os.path.join(writer.build_directory, "c")
            assert os.path.isfile(dest)
            with open(dest) as f:
                output = f.read()
                self.assertEqual(output, "d")

    def test_relpath_precedence(self):
        """Does the FilesWriter relpath option take precedence over the path?"""

        # Work in a temporary directory.
        with self.create_temp_cwd():

            # Create test file
            os.mkdir("sub")
            with open(os.path.join("sub", "c"), "w") as f:
                f.write("d")

            # Create the resoruces dictionary
            res = dict(metadata=dict(path="other_sub"))

            # Create files writer, test output
            writer = FilesWriter()
            writer.files = [os.path.join("sub", "c")]
            writer.build_directory = "build"
            writer.relpath = "sub"
            writer.write("y", res, notebook_name="z")

            # Check the output of the file
            assert os.path.isdir(writer.build_directory)
            dest = os.path.join(writer.build_directory, "z")
            with open(dest) as f:
                output = f.read()
                self.assertEqual(output, "y")

            # Check to make sure the linked file was copied
            dest = os.path.join(writer.build_directory, "c")
            assert os.path.isfile(dest)
            with open(dest) as f:
                output = f.read()
                self.assertEqual(output, "d")



"""Tests for MarkdownExporter"""

# -----------------------------------------------------------------------------
# Copyright (c) 2013, the IPython Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

from ..markdown import MarkdownExporter
from .base import ExportersTestsBase

# -----------------------------------------------------------------------------
# Class
# -----------------------------------------------------------------------------


class TestMarkdownExporter(ExportersTestsBase):
    """Tests for MarkdownExporter"""

    exporter_class = MarkdownExporter
    should_include_raw = ["markdown", "html"]

    def test_constructor(self):
        """
        Can a MarkdownExporter be constructed?
        """
        MarkdownExporter()

    def test_export(self):
        """
        Can a MarkdownExporter export something?
        """
        (output, resources) = MarkdownExporter().from_filename(self._get_notebook())
        assert len(output) > 0



"""
Module with tests for export.py
"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import os
import sys

import nbformat
import pytest
from traitlets.config import Config

import nbconvert.tests

from ..base import (
    ExporterDisabledError,
    ExporterNameError,
    export,
    get_export_names,
    get_exporter,
)
from ..exporter import Exporter
from ..python import PythonExporter
from .base import ExportersTestsBase


class TestExport(ExportersTestsBase):
    """Contains test functions for export.py"""

    def test_export_wrong_name(self):
        """
        Is the right error thrown when a bad template name is used?
        """
        try:
            exporter = get_exporter("not_a_name")
            export(exporter, self._get_notebook())
        except ExporterNameError as e:
            pass

    def test_export_disabled(self):
        """
        Trying to use a disabled exporter should raise ExporterDisbledError
        """
        config = Config({"NotebookExporter": {"enabled": False}})
        with pytest.raises(ExporterDisabledError):
            get_exporter("notebook", config=config)

    def test_export_filename(self):
        """
        Can a notebook be exported by filename?
        """
        exporter = get_exporter("python")
        (output, resources) = export(exporter, self._get_notebook())
        assert len(output) > 0

    def test_export_nbnode(self):
        """
        Can a notebook be exported by a notebook node handle?
        """
        with open(self._get_notebook()) as f:
            notebook = nbformat.read(f, 4)
            exporter = get_exporter("python")
            (output, resources) = export(exporter, notebook)
        assert len(output) > 0

    def test_export_filestream(self):
        """
        Can a notebook be exported by a filesteam?
        """
        with open(self._get_notebook()) as f:
            exporter = get_exporter("python")
            (output, resources) = export(exporter, f)
        assert len(output) > 0

    def test_export_using_exporter(self):
        """
        Can a notebook be exported using an instanciated exporter?
        """
        (output, resources) = export(PythonExporter(), self._get_notebook())
        assert len(output) > 0

    def test_export_using_exporter_class(self):
        """
        Can a notebook be exported using an exporter class type?
        """
        (output, resources) = export(PythonExporter, self._get_notebook())
        assert len(output) > 0

    def test_export_resources(self):
        """
        Can a notebook be exported along with a custom resources dict?
        """
        (output, resources) = export(PythonExporter, self._get_notebook(), resources={})
        assert len(output) > 0

    def test_no_exporter(self):
        """
        Is the right error thrown if no exporter is provided?
        """
        try:
            (output, resources) = export(None, self._get_notebook())
        except TypeError:
            pass


def test_get_exporter_entrypoint():
    p = os.path.join(os.path.dirname(nbconvert.tests.__file__), "exporter_entrypoint")
    sys.path.insert(0, p)
    assert "entrypoint_test" in get_export_names()
    try:
        cls = get_exporter("entrypoint_test")
        assert issubclass(cls, Exporter), cls
    finally:
        del sys.path[0]



"""Tests for the latex preprocessor"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

from unittest.mock import patch

import pytest

from ..webpdf import WebPDFExporter
from .base import ExportersTestsBase


class TestWebPDFExporter(ExportersTestsBase):
    """Contains test functions for webpdf.py"""

    exporter_class = WebPDFExporter

    @pytest.mark.network
    def test_export(self):
        """
        Can a TemplateExporter export something?
        """
        (output, resources) = WebPDFExporter(allow_chromium_download=True).from_filename(
            self._get_notebook()
        )
        assert len(output) > 0

    @patch("pyppeteer.util.check_chromium", return_value=False)
    def test_webpdf_without_chromium(self, mock_check_chromium):
        """
        Generate PDFs if chromium not present?
        """
        with pytest.raises(RuntimeError):
            WebPDFExporter(allow_chromium_download=False).from_filename(self._get_notebook())

    def test_webpdf_without_pyppeteer(self):
        """
        Generate PDFs if chromium not present?
        """
        with pytest.raises(RuntimeError):
            exporter = WebPDFExporter()
            with open(self._get_notebook(), encoding="utf-8") as f:
                nb = exporter.from_file(f, resources={})
                # Have to do this as the very last action as traitlets do dynamic importing often
                with patch("builtins.__import__", side_effect=ModuleNotFoundError("Fake missing")):
                    exporter.from_notebook_node(nb)



"""Tests for ASCIIDocExporter`"""

# -----------------------------------------------------------------------------
# Copyright (c) 2016, the IPython Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

import re

from traitlets.config import Config

from ...tests.utils import onlyif_cmds_exist
from ..asciidoc import ASCIIDocExporter
from .base import ExportersTestsBase

# -----------------------------------------------------------------------------
# Class
# -----------------------------------------------------------------------------
in_regex = r"In\[(.*)\]:"
out_regex = r"Out\[(.*)\]:"


class TestASCIIDocExporter(ExportersTestsBase):
    """Tests for ASCIIDocExporter"""

    exporter_class = ASCIIDocExporter

    def test_constructor(self):
        """
        Can a ASCIIDocExporter be constructed?
        """
        ASCIIDocExporter()

    @onlyif_cmds_exist("pandoc")
    def test_export(self):
        """
        Can a ASCIIDocExporter export something?
        """
        (output, resources) = ASCIIDocExporter().from_filename(self._get_notebook())
        assert len(output) > 0

        assert re.findall(in_regex, output)
        assert re.findall(out_regex, output)

    @onlyif_cmds_exist("pandoc")
    def test_export_no_prompt(self):
        """
        Can a ASCIIDocExporter export something without prompts?
        """
        no_prompt = {
            "TemplateExporter": {
                "exclude_input_prompt": True,
                "exclude_output_prompt": True,
            }
        }
        c_no_prompt = Config(no_prompt)
        exporter = ASCIIDocExporter(config=c_no_prompt)
        (output, resources) = exporter.from_filename(
            self._get_notebook(nb_name="prompt_numbers.ipynb")
        )

        assert not re.findall(in_regex, output)
        assert not re.findall(out_regex, output)



from .asciidoc import ASCIIDocExporter
from .base import ExporterNameError, export, get_export_names, get_exporter
from .exporter import Exporter, FilenameExtension
from .html import HTMLExporter
from .latex import LatexExporter
from .markdown import MarkdownExporter
from .notebook import NotebookExporter
from .pdf import PDFExporter
from .python import PythonExporter
from .rst import RSTExporter
from .script import ScriptExporter
from .slides import SlidesExporter
from .templateexporter import TemplateExporter
from .webpdf import WebPDFExporter



"""Preprocessor for merging consecutive stream outputs for easier handling."""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import functools
import re

from traitlets.log import get_logger


def cell_preprocessor(function):
    """
    Wrap a function to be executed on all cells of a notebook

    The wrapped function should have these parameters:

    cell : NotebookNode cell
        Notebook cell being processed
    resources : dictionary
        Additional resources used in the conversion process.  Allows
        preprocessors to pass variables into the Jinja engine.
    index : int
        Index of the cell being processed
    """

    @functools.wraps(function)
    def wrappedfunc(nb, resources):
        get_logger().debug("Applying preprocessor: %s", function.__name__)
        for index, cell in enumerate(nb.cells):
            nb.cells[index], resources = function(cell, resources, index)
        return nb, resources

    return wrappedfunc


cr_pat = re.compile(r".*\r(?=[^\n])")


@cell_preprocessor
def coalesce_streams(cell, resources, index):
    """
    Merge consecutive sequences of stream output into single stream
    to prevent extra newlines inserted at flush calls

    Parameters
    ----------
    cell : NotebookNode cell
        Notebook cell being processed
    resources : dictionary
        Additional resources used in the conversion process.  Allows
        transformers to pass variables into the Jinja engine.
    index : int
        Index of the cell being processed
    """

    outputs = cell.get("outputs", [])
    if not outputs:
        return cell, resources

    last = outputs[0]
    new_outputs = [last]
    for output in outputs[1:]:
        if (
            output.output_type == "stream"
            and last.output_type == "stream"
            and last.name == output.name
        ):
            last.text += output.text

        else:
            new_outputs.append(output)
            last = output

    # process \r characters
    for output in new_outputs:
        if output.output_type == "stream" and "\r" in output.text:
            output.text = cr_pat.sub("", output.text)

    cell.outputs = new_outputs
    return cell, resources



"""
Module containing filter functions that allow code to be highlighted
from within Jinja templates.
"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

# pygments must not be imported at the module level
# because errors should be raised at runtime if it's actually needed,
# not import time, when it may not be needed.

from html import escape
from warnings import warn

from traitlets import observe
from traitlets.config import Dict

from nbconvert.utils.base import NbConvertBase

MULTILINE_OUTPUTS = ["text", "html", "svg", "latex", "javascript", "json"]

__all__ = ["Highlight2HTML", "Highlight2Latex"]


class Highlight2HTML(NbConvertBase):
    extra_formatter_options = Dict(
        {},
        help="""
        Extra set of options to control how code is highlighted.

        Passed through to the pygments' HtmlFormatter class.
        See available list in https://pygments.org/docs/formatters/#HtmlFormatter
        """,
        config=True,
    )

    def __init__(self, pygments_lexer=None, **kwargs):
        self.pygments_lexer = pygments_lexer or "ipython3"
        super().__init__(**kwargs)

    @observe("default_language")
    def _default_language_changed(self, change):
        warn(
            "Setting default_language in config is deprecated as of 5.0, "
            "please use language_info metadata instead."
        )
        self.pygments_lexer = change["new"]

    def __call__(self, source, language=None, metadata=None):
        """
        Return a syntax-highlighted version of the input source as html output.

        Parameters
        ----------
        source : str
            source of the cell to highlight
        language : str
            language to highlight the syntax of
        metadata : NotebookNode cell metadata
            metadata of the cell to highlight
        """
        from pygments.formatters import HtmlFormatter

        if not language:
            language = self.pygments_lexer

        return _pygments_highlight(
            source if len(source) > 0 else " ",
            # needed to help post processors:
            HtmlFormatter(
                cssclass=escape(f" highlight hl-{language}"), **self.extra_formatter_options
            ),
            language,
            metadata,
        )


class Highlight2Latex(NbConvertBase):
    extra_formatter_options = Dict(
        {},
        help="""
        Extra set of options to control how code is highlighted.

        Passed through to the pygments' LatexFormatter class.
        See available list in https://pygments.org/docs/formatters/#LatexFormatter
        """,
        config=True,
    )

    def __init__(self, pygments_lexer=None, **kwargs):
        self.pygments_lexer = pygments_lexer or "ipython3"
        super().__init__(**kwargs)

    @observe("default_language")
    def _default_language_changed(self, change):
        warn(
            "Setting default_language in config is deprecated as of 5.0, "
            "please use language_info metadata instead."
        )
        self.pygments_lexer = change["new"]

    def __call__(self, source, language=None, metadata=None, strip_verbatim=False):
        """
        Return a syntax-highlighted version of the input source as latex output.

        Parameters
        ----------
        source : str
            source of the cell to highlight
        language : str
            language to highlight the syntax of
        metadata : NotebookNode cell metadata
            metadata of the cell to highlight
        strip_verbatim : bool
            remove the Verbatim environment that pygments provides by default
        """
        from pygments.formatters import LatexFormatter

        if not language:
            language = self.pygments_lexer

        latex = _pygments_highlight(
            source, LatexFormatter(**self.extra_formatter_options), language, metadata
        )
        if strip_verbatim:
            latex = latex.replace(r"\begin{Verbatim}[commandchars=\\\{\}]" + "\n", "")  # noqa
            return latex.replace("\n\\end{Verbatim}\n", "")
        else:
            return latex


def _pygments_highlight(source, output_formatter, language="ipython", metadata=None):
    """
    Return a syntax-highlighted version of the input source

    Parameters
    ----------
    source : str
        source of the cell to highlight
    output_formatter : Pygments formatter
    language : str
        language to highlight the syntax of
    metadata : NotebookNode cell metadata
        metadata of the cell to highlight
    """
    from pygments import highlight
    from pygments.lexers import get_lexer_by_name
    from pygments.util import ClassNotFound

    # If the cell uses a magic extension language,
    # use the magic language instead.
    if language.startswith("ipython") and metadata and "magics_language" in metadata:

        language = metadata["magics_language"]

    lexer = None
    if language == "ipython2":
        try:
            from IPython.lib.lexers import IPythonLexer
        except ImportError:
            warn("IPython lexer unavailable, falling back on Python")
            language = "python"
        else:
            lexer = IPythonLexer()
    elif language == "ipython3":
        try:
            from IPython.lib.lexers import IPython3Lexer
        except ImportError:
            warn("IPython3 lexer unavailable, falling back on Python 3")
            language = "python3"
        else:
            lexer = IPython3Lexer()

    if lexer is None:
        try:
            lexer = get_lexer_by_name(language, stripall=True)
        except ClassNotFound:
            warn("No lexer found for language %r. Treating as plain text." % language)
            from pygments.lexers.special import TextLexer

            lexer = TextLexer()

    return highlight(source, lexer, output_formatter)



"""
Module with tests for the clearmetadata preprocessor.
"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

from ..clearmetadata import ClearMetadataPreprocessor
from .base import PreprocessorTestsBase


class TestClearMetadata(PreprocessorTestsBase):
    """Contains test functions for clearmetadata.py"""

    def build_notebook(self):
        notebook = super().build_notebook()
        notebook.metadata = {
            "language_info": {"name": "python", "version": "3.6.7"},
            "kernelspec": {"language": "python", "name": "python3"},
        }
        # Add a test field to the first cell
        if "metadata" not in notebook.cells[0]:
            notebook.cells[0].metadata = {}
        notebook.cells[0].metadata["test_field"] = "test_value"
        notebook.cells[0].metadata["test_nested"] = {"test_keep": "keep", "test_filtered": "filter"}
        notebook.cells[0].metadata["executeTime"] = dict(
            [("end_time", "09:31:50"), ("start_time", "09:31:49")]
        )
        return notebook

    def build_preprocessor(self, **kwargs):
        """Make an instance of a preprocessor"""
        preprocessor = ClearMetadataPreprocessor(**kwargs)
        preprocessor.enabled = True
        return preprocessor

    def test_constructor(self):
        """Can a ClearMetadataPreprocessor be constructed?"""
        self.build_preprocessor()

    def test_default_output(self):
        """Test the output of the ClearMetadataPreprocessor"""
        nb = self.build_notebook()
        res = self.build_resources()
        preprocessor = self.build_preprocessor()
        nb, res = preprocessor(nb, res)

        assert not nb.cells[0].metadata
        # By default we only perserve the langauge name
        assert nb.metadata == {"language_info": {"name": "python"}}

    def test_cell_only(self):
        """Test the output of the ClearMetadataPreprocessor"""
        nb = self.build_notebook()
        res = self.build_resources()
        preprocessor = self.build_preprocessor(clear_notebook_metadata=False)
        nb, res = preprocessor(nb, res)

        assert not nb.cells[0].metadata
        assert nb.metadata

    def test_notebook_only(self):
        """Test the output of the ClearMetadataPreprocessor"""
        nb = self.build_notebook()
        res = self.build_resources()
        preprocessor = self.build_preprocessor(
            clear_cell_metadata=False, preserve_nb_metadata_mask=set()
        )
        nb, res = preprocessor(nb, res)

        assert nb.cells[0].metadata
        assert not nb.metadata

    def test_selective_cell_metadata(self):
        """Test the output of the ClearMetadataPreprocessor"""
        nb = self.build_notebook()
        res = self.build_resources()
        preprocessor = self.build_preprocessor(
            preserve_cell_metadata_mask=["test_field"], preserve_nb_metadata_mask=set()
        )
        nb, res = preprocessor(nb, res)

        assert nb.cells[0].metadata == {"test_field": "test_value"}
        assert not nb.metadata

    def test_selective_cell_tuple_metadata(self):
        """Test the output of the ClearMetadataPreprocessor"""
        nb = self.build_notebook()
        res = self.build_resources()
        # Ensure that a tuple length 1 works as well as a string key
        preprocessor = self.build_preprocessor(
            preserve_cell_metadata_mask=[("test_field",)], preserve_nb_metadata_mask=set()
        )
        nb, res = preprocessor(nb, res)

        assert nb.cells[0].metadata == {"test_field": "test_value"}
        assert not nb.metadata

    def test_nested_cell_metadata(self):
        """Test the output of the ClearMetadataPreprocessor"""
        nb = self.build_notebook()
        res = self.build_resources()
        preprocessor = self.build_preprocessor(
            preserve_cell_metadata_mask=[("test_nested", "test_keep")],
            preserve_nb_metadata_mask=set(),
        )
        nb, res = preprocessor(nb, res)

        assert nb.cells[0].metadata == {"test_nested": {"test_keep": "keep"}}
        assert not nb.metadata

    def test_nested_cell_tuple_metadata(self):
        """Test the output of the ClearMetadataPreprocessor"""
        nb = self.build_notebook()
        res = self.build_resources()
        # Ensure that a tuple length 1 works as well as a string key
        preprocessor = self.build_preprocessor(
            preserve_cell_metadata_mask=[("test_nested", ("test_keep",))],
            preserve_nb_metadata_mask=set(),
        )
        nb, res = preprocessor(nb, res)

        assert nb.cells[0].metadata == {"test_nested": {"test_keep": "keep"}}
        assert not nb.metadata

    def test_selective_notebook_metadata(self):
        """Test the output of the ClearMetadataPreprocessor"""
        nb = self.build_notebook()
        res = self.build_resources()
        preprocessor = self.build_preprocessor(preserve_nb_metadata_mask=["kernelspec"])
        nb, res = preprocessor(nb, res)

        assert not nb.cells[0].metadata
        assert nb.metadata == {"kernelspec": {"language": "python", "name": "python3"}}



"""Tests for notebook.py"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import json

from nbformat import validate

from nbconvert.tests.base import assert_big_text_equal

from ..notebook import NotebookExporter
from .base import ExportersTestsBase


class TestNotebookExporter(ExportersTestsBase):
    """Contains test functions for notebook.py"""

    exporter_class = NotebookExporter

    def test_export(self):
        """
        Does the NotebookExporter return the file unchanged?
        """
        with open(self._get_notebook()) as f:
            file_contents = f.read()
        (output, resources) = self.exporter_class().from_filename(self._get_notebook())
        assert len(output) > 0
        assert_big_text_equal(output, file_contents)

    def test_downgrade_3(self):
        exporter = self.exporter_class(nbformat_version=3)
        (output, resources) = exporter.from_filename(self._get_notebook())
        nb = json.loads(output)
        validate(nb)

    def test_downgrade_2(self):
        exporter = self.exporter_class(nbformat_version=2)
        (output, resources) = exporter.from_filename(self._get_notebook())
        nb = json.loads(output)
        self.assertEqual(nb["nbformat"], 2)



"""Tests for the extractoutput preprocessor"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import json

from ..extractoutput import ExtractOutputPreprocessor
from .base import PreprocessorTestsBase


class TestExtractOutput(PreprocessorTestsBase):
    """Contains test functions for extractoutput.py"""

    def build_preprocessor(self):
        """Make an instance of a preprocessor"""
        preprocessor = ExtractOutputPreprocessor()
        preprocessor.extract_output_types = {"text/plain", "image/png", "application/pdf"}
        preprocessor.enabled = True
        return preprocessor

    def test_constructor(self):
        """Can a ExtractOutputPreprocessor be constructed?"""
        self.build_preprocessor()

    def test_output(self):
        """Test the output of the ExtractOutputPreprocessor"""
        nb = self.build_notebook()
        res = self.build_resources()
        preprocessor = self.build_preprocessor()
        nb, res = preprocessor(nb, res)
        # Check if text was extracted.
        output = nb.cells[0].outputs[1]
        self.assertIn("filenames", output.metadata)
        self.assertIn("text/plain", output.metadata.filenames)
        text_filename = output.metadata.filenames["text/plain"]

        # Check if png was extracted.
        output = nb.cells[0].outputs[6]
        self.assertIn("filenames", output.metadata)
        self.assertIn("image/png", output.metadata.filenames)
        png_filename = output.metadata.filenames["image/png"]

        # Check that pdf was extracted
        output = nb.cells[0].outputs[7]
        self.assertIn("filenames", output.metadata)
        self.assertIn("application/pdf", output.metadata.filenames)
        pdf_filename = output.metadata.filenames["application/pdf"]

        # Verify text output
        self.assertIn(text_filename, res["outputs"])
        self.assertEqual(res["outputs"][text_filename], b"b")

        # Verify png output
        self.assertIn(png_filename, res["outputs"])
        self.assertEqual(res["outputs"][png_filename], b"g")

        # Verify pdf output
        self.assertIn(pdf_filename, res["outputs"])
        self.assertEqual(res["outputs"][pdf_filename], b"h")

    def test_json_extraction(self):
        nb = self.build_notebook(with_json_outputs=True)
        res = self.build_resources()
        preprocessor = self.build_preprocessor()
        preprocessor.extract_output_types = {"application/json"}
        nb, res = preprocessor(nb, res)
        reference = self.build_notebook(with_json_outputs=True).cells[0].outputs

        # Verify cell untouched
        self.assertEqual(
            [out.get("data") for out in nb.cells[0].outputs], [out.get("data") for out in reference]
        )

        outputs = sorted(res["outputs"].values())
        reference_files = []
        for out in reference:
            try:
                data = out["data"]["application/json"]
                reference_files.append(json.dumps(data).encode())
            except KeyError:
                pass

        # Verify equivalence of extracted outputs.
        self.assertEqual(sorted(outputs), sorted(reference_files))



"""Export to PDF via latex"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import os
import shutil
import subprocess
import sys
from tempfile import TemporaryDirectory

from traitlets import Bool, Instance, Integer, List, Unicode, default

from ..utils import _contextlib_chdir
from .latex import LatexExporter


class LatexFailed(IOError):
    """Exception for failed latex run

    Captured latex output is in error.output.
    """

    def __init__(self, output):
        self.output = output

    def __unicode__(self):
        return "PDF creating failed, captured latex output:\n%s" % self.output

    def __str__(self):
        u = self.__unicode__()
        return u


def prepend_to_env_search_path(varname, value, envdict):
    """Add value to the environment variable varname in envdict

    e.g. prepend_to_env_search_path('BIBINPUTS', '/home/sally/foo', os.environ)
    """
    if not value:
        return  # Nothing to add

    envdict[varname] = value + os.pathsep + envdict.get(varname, "")


class PDFExporter(LatexExporter):
    """Writer designed to write to PDF files.

    This inherits from `LatexExporter`. It creates a LaTeX file in
    a temporary directory using the template machinery, and then runs LaTeX
    to create a pdf.
    """

    export_from_notebook = "PDF via LaTeX"

    latex_count = Integer(3, help="How many times latex will be called.").tag(config=True)

    latex_command = List(
        ["xelatex", "{filename}", "-quiet"], help="Shell command used to compile latex."
    ).tag(config=True)

    bib_command = List(["bibtex", "{filename}"], help="Shell command used to run bibtex.").tag(
        config=True
    )

    verbose = Bool(False, help="Whether to display the output of latex commands.").tag(config=True)

    texinputs = Unicode(help="texinputs dir. A notebook's directory is added")
    writer = Instance("nbconvert.writers.FilesWriter", args=(), kw={"build_directory": "."})

    output_mimetype = "application/pdf"

    _captured_output = List()

    @default("file_extension")
    def _file_extension_default(self):
        return ".pdf"

    @default("template_extension")
    def _template_extension_default(self):
        return ".tex.j2"

    def run_command(self, command_list, filename, count, log_function, raise_on_failure=None):
        """Run command_list count times.

        Parameters
        ----------
        command_list : list
            A list of args to provide to Popen. Each element of this
            list will be interpolated with the filename to convert.
        filename : unicode
            The name of the file to convert.
        count : int
            How many times to run the command.
        raise_on_failure: Exception class (default None)
            If provided, will raise the given exception for if an instead of
            returning False on command failure.

        Returns
        -------
        success : bool
            A boolean indicating if the command was successful (True)
            or failed (False).
        """
        command = [c.format(filename=filename) for c in command_list]

        # This will throw a clearer error if the command is not found
        cmd = shutil.which(command_list[0])
        if cmd is None:
            link = "https://nbconvert.readthedocs.io/en/latest/install.html#installing-tex"
            raise OSError(
                "{formatter} not found on PATH, if you have not installed "
                "{formatter} you may need to do so. Find further instructions "
                "at {link}.".format(formatter=command_list[0], link=link)
            )

        times = "time" if count == 1 else "times"
        self.log.info("Running %s %i %s: %s", command_list[0], count, times, command)

        shell = sys.platform == "win32"
        if shell:
            command = subprocess.list2cmdline(command)
        env = os.environ.copy()
        prepend_to_env_search_path("TEXINPUTS", self.texinputs, env)
        prepend_to_env_search_path("BIBINPUTS", self.texinputs, env)
        prepend_to_env_search_path("BSTINPUTS", self.texinputs, env)

        with open(os.devnull, "rb") as null:
            stdout = subprocess.PIPE if not self.verbose else None
            for _ in range(count):
                p = subprocess.Popen(
                    command,
                    stdout=stdout,
                    stderr=subprocess.STDOUT,
                    stdin=null,
                    shell=shell,
                    env=env,
                )
                out, _ = p.communicate()
                if p.returncode:
                    if self.verbose:
                        # verbose means I didn't capture stdout with PIPE,
                        # so it's already been displayed and `out` is None.
                        out = ""
                    else:
                        out = out.decode("utf-8", "replace")
                    log_function(command, out)
                    self._captured_output.append(out)
                    if raise_on_failure:
                        raise raise_on_failure(
                            'Failed to run "{command}" command:\n{output}'.format(
                                command=command, output=out
                            )
                        )
                    return False  # failure
        return True  # success

    def run_latex(self, filename, raise_on_failure=LatexFailed):
        """Run xelatex self.latex_count times."""

        def log_error(command, out):
            self.log.critical("%s failed: %s\n%s", command[0], command, out)

        return self.run_command(
            self.latex_command, filename, self.latex_count, log_error, raise_on_failure
        )

    def run_bib(self, filename, raise_on_failure=False):
        """Run bibtex one time."""
        filename = os.path.splitext(filename)[0]

        def log_error(command, out):
            self.log.warning(
                "%s had problems, most likely because there were no citations", command[0]
            )
            self.log.debug("%s output: %s\n%s", command[0], command, out)

        return self.run_command(self.bib_command, filename, 1, log_error, raise_on_failure)

    def from_notebook_node(self, nb, resources=None, **kw):
        latex, resources = super().from_notebook_node(nb, resources=resources, **kw)
        # set texinputs directory, so that local files will be found
        if resources and resources.get("metadata", {}).get("path"):
            self.texinputs = resources["metadata"]["path"]
        else:
            self.texinputs = os.getcwd()

        self._captured_outputs = []
        with TemporaryDirectory() as td, _contextlib_chdir.chdir(td):
            notebook_name = "notebook"
            resources["output_extension"] = ".tex"
            tex_file = self.writer.write(latex, resources, notebook_name=notebook_name)
            self.log.info("Building PDF")
            self.run_latex(tex_file)
            if self.run_bib(tex_file):
                self.run_latex(tex_file)

            pdf_file = notebook_name + ".pdf"
            if not os.path.isfile(pdf_file):
                raise LatexFailed("\n".join(self._captured_output))
            self.log.info("PDF successfully created")
            with open(pdf_file, "rb") as f:
                pdf_data = f.read()

        # convert output extension to pdf
        # the writer above required it to be tex
        resources["output_extension"] = ".pdf"
        # clear figure outputs, extracted by latex export,
        # so we don't claim to be a multi-file export.
        resources.pop("outputs", None)

        return pdf_data, resources



"""Base TestCase class for testing Exporters"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import os

from ...tests.base import TestsBase

all_raw_mimetypes = {
    "text/x-python",
    "text/markdown",
    "text/html",
    "text/restructuredtext",
    "text/latex",
}


class ExportersTestsBase(TestsBase):
    """Contains base test functions for exporters"""

    exporter_class = None
    should_include_raw = None

    def _get_notebook(self, nb_name="notebook2.ipynb"):
        return os.path.join(self._get_files_path(), nb_name)

    def test_raw_cell_inclusion(self):
        """test raw cell inclusion based on raw_mimetype metadata"""
        if self.should_include_raw is None:
            return
        exporter = self.exporter_class()
        (output, resources) = exporter.from_filename(self._get_notebook("rawtest.ipynb"))
        for inc in self.should_include_raw:
            self.assertIn("raw %s" % inc, output, "should include %s" % inc)
        self.assertIn("no raw_mimetype metadata", output)
        for exc in all_raw_mimetypes.difference(self.should_include_raw):
            self.assertNotIn("raw %s" % exc, output, "should exclude %s" % exc)
        self.assertNotIn("never be included", output)



"""LaTeX Exporter class"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.


from traitlets import default
from traitlets.config import Config

from nbconvert.filters.filter_links import resolve_references
from nbconvert.filters.highlight import Highlight2Latex

from .templateexporter import TemplateExporter


class LatexExporter(TemplateExporter):
    """
    Exports to a Latex template.  Inherit from this class if your template is
    LaTeX based and you need custom transformers/filters.
    If you don't need custom transformers/filters, just change the
    'template_file' config option.  Place your template in the special "/latex"
    subfolder of the "../templates" folder.
    """

    export_from_notebook = "LaTeX"

    @default("file_extension")
    def _file_extension_default(self):
        return ".tex"

    @default("template_name")
    def _template_name_default(self):
        return "latex"

    output_mimetype = "text/latex"

    def default_filters(self):
        yield from super().default_filters()
        yield ("resolve_references", resolve_references)

    @property
    def default_config(self):
        c = Config(
            {
                "NbConvertBase": {
                    "display_data_priority": [
                        "text/latex",
                        "application/pdf",
                        "image/png",
                        "image/jpeg",
                        "image/svg+xml",
                        "text/markdown",
                        "text/plain",
                    ]
                },
                "ExtractOutputPreprocessor": {"enabled": True},
                "SVG2PDFPreprocessor": {"enabled": True},
                "LatexPreprocessor": {"enabled": True},
                "SphinxPreprocessor": {"enabled": True},
                "HighlightMagicsPreprocessor": {"enabled": True},
            }
        )
        c.merge(super().default_config)
        return c

    def from_notebook_node(self, nb, resources=None, **kw):
        langinfo = nb.metadata.get("language_info", {})
        lexer = langinfo.get("pygments_lexer", langinfo.get("name", None))
        highlight_code = self.filters.get(
            "highlight_code", Highlight2Latex(pygments_lexer=lexer, parent=self)
        )
        self.register_filter("highlight_code", highlight_code)

        return super().from_notebook_node(nb, resources, **kw)

    def _create_environment(self):
        environment = super()._create_environment()

        # Set special Jinja2 syntax that will not conflict with latex.
        environment.block_start_string = "((*"
        environment.block_end_string = "*))"
        environment.variable_start_string = "((("
        environment.variable_end_string = ")))"
        environment.comment_start_string = "((="
        environment.comment_end_string = "=))"

        return environment



"""Module containing a preprocessor that converts outputs in the notebook from
one format to another.
"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.


from traitlets import Unicode

from .base import Preprocessor


class ConvertFiguresPreprocessor(Preprocessor):
    """
    Converts all of the outputs in a notebook from one format to another.
    """

    from_format = Unicode(help="Format the converter accepts").tag(config=True)
    to_format = Unicode(help="Format the converter writes").tag(config=True)

    def __init__(self, **kw):
        """
        Public constructor
        """
        super().__init__(**kw)

    def convert_figure(self, data_format, data):
        raise NotImplementedError()

    def preprocess_cell(self, cell, resources, cell_index):
        """
        Apply a transformation on each cell,

        See base.py
        """

        # Loop through all of the datatypes of the outputs in the cell.
        for output in cell.get("outputs", []):
            if (
                output.output_type in {"execute_result", "display_data"}
                and self.from_format in output.data
                and self.to_format not in output.data
            ):

                output.data[self.to_format] = self.convert_figure(
                    self.from_format, output.data[self.from_format]
                )

        return cell, resources
